from typing import Any, List, Optional, Union

import numpy as np
import pandas as pd
import planetary_computer as pc
import pystac_client
import rasterio.features
import stackstac
import xarray as xr
from scipy import constants

from .utils import _central_pixel_bbox, _compute_distance_to_center


def create(
    lat: Union[float, int],
    lon: Union[float, int],
    collection: str,
    start_date: str,
    end_date: str,
    bands: Optional[Union[str, List[str]]] = None,
    edge_size: Union[float, int] = 128.0,
    units: str = "px",
    resolution: Union[float, int] = 10.0,
    stac: str = "https://planetarycomputer.microsoft.com/api/stac/v1",
    gee: bool = False,
    stackstac_kw: Optional[dict] = None,
    **kwargs,
) -> xr.DataArray:
    """Creates a data cube from a STAC Catalogue as a :code:`xr.DataArray` object.

    The coordinates used here work as the approximate central coordinates of the data cube
    in the spatial dimension.

    Parameters
    ----------
    lat : float | int
        Latitude of the central pixel of the data cube.
    lon : float | int
        Longitude of the central pixel of the data cube.
    collection : str
        Name of the collection in the STAC Catalogue.
    start_date : str
        Start date of the data cube in YYYY-MM-DD format.
    end_date : str
        End date of the data cube in YYYY-MM-DD format.
    bands : str | List[str], default = None
        Name of the band(s) from the collection to use.
    edge_size : float | int, default = 128
        Size of the edge of the cube in the units specified by :code:`units`. All edges share the same size.

        .. warning::
           If :code:`edge_size` is not a multiple of 2, it will be rounded.

    units : str, default = 'px'
        Units of the provided edge size in :code:`edge_size`. Must be 'px' (pixels), 'm' (meters), or a unit
        name in https://docs.scipy.org/doc/scipy/reference/constants.html#units.

        .. versionadded:: 2024.1.1

        .. warning::
           Note that when :code:`units!='px'` the edge size will be transformed to meters (if :code:`units!='m'`).
           Furthermore, the edge size will be converted to pixels (using :code:`edge_size/resolution`)
           and rounded (see :code:`edge_size`). Therefore, the edge size when :code:`units!='px'` is just an approximation
           if its value in meters is not divisible by the requested resolution.

    resolution : float | int, default = 10
        Pixel size in meters.
    stac : str, default = 'https://planetarycomputer.microsoft.com/api/stac/v1'
        Endpoint of the STAC Catalogue to use.
    gee : bool, default = True
        Whether to use Google Earth Engine. This ignores the 'stac' argument.

        .. versionadded:: 2024.1.0

    stackstac_kw : dict, default = None
        Keyword arguments for :code:`stackstac` as a dictionary.

        .. versionadded:: 2024.1.0

    kwargs :
        Additional keyword arguments passed to :code:`pystac_client.Client.search()`.

    Returns
    -------
    xr.DataArray
        Data Cube.


    Examples
    --------
    Create a Sentinel-2 L2A data cube with an edge size of 64 px from Planetary Computer:

    >>> import cubo
    >>> cubo.create(
    ...     lat=50,
    ...     lon=10,
    ...     collection="sentinel-2-l2a",
    ...     bands=["B02","B03","B04"],
    ...     start_date="2021-06-01",
    ...     end_date="2021-06-10",
    ...     edge_size=32,
    ...     resolution=10,
    ... )
    <xarray.DataArray (time: 3, band: 3, x: 32, y: 32)>

    Create a Sentinel-2 L2A data cube with an edge size of 128 px from Google Earth Engine:

    >>> import cubo
    >>> cubo.create(
    ...     lat=51.079225,
    ...     lon=10.452173,
    ...     collection="COPERNICUS/S2_SR_HARMONIZED",
    ...     bands=["B2","B3","B4"],
    ...     start_date="2016-06-01",
    ...     end_date="2017-07-01",
    ...     edge_size=128,
    ...     resolution=10,
    ...     gee=True,
    ... )
    <xarray.DataArray (time: 27, band: 3, x: 128, y: 128)>
    """
    # Harmonize units to pixels
    if units != "px":
        if units == "m":
            edge_size = edge_size / resolution
        else:
            edge_size = (edge_size * getattr(constants, units)) / resolution

    # Get the BBox and EPSG
    bbox_utm, bbox_latlon, utm_coords, epsg = _central_pixel_bbox(
        lat, lon, edge_size, resolution
    )

    # Use Google Earth Engine
    if gee:

        # Try to import ee, otherwise raise an ImportError
        try:
            import ee
            import xee
        except ImportError:
            raise ImportError(
                '"earthengine-api" and "xee" could not be loaded. Please install them, or install "cubo" using "pip install cubo[ee]"'
            )

        # Initialize Google Earth Engine with the high volume endpoint
        # User must do this before using cubo
        # ee.Initialize(opt_url="https://earthengine-highvolume.googleapis.com")

        # Get BBox values in latlon
        west = bbox_latlon["coordinates"][0][0][0]
        south = bbox_latlon["coordinates"][0][0][1]
        east = bbox_latlon["coordinates"][0][2][0]
        north = bbox_latlon["coordinates"][0][2][1]

        # Create the BBox geometry in GEE
        BBox = ee.Geometry.BBox(west, south, east, north)

        # If the collection is string then access the Image Collection
        if isinstance(collection, str):
            collection = ee.ImageCollection(collection)

        # Do the filtering: Bounds, time, and bands
        collection = (
            collection.filterBounds(BBox).filterDate(start_date, end_date).select(bands)
        )

        # Return the cube via xee
        cube = xr.open_dataset(
            collection,
            engine="ee",
            geometry=BBox,
            scale=resolution,
            crs=f"EPSG:{epsg}",
            chunks=dict(),
        )

        # Rename the coords to match stackstac names, also rearrange
        cube = (
            cube.rename(Y="y", X="x")
            .to_array("band")
            .transpose("time", "band", "y", "x")
        )

        # Delete all attributes
        cube.attrs = dict()

        # Get the name of the collection
        collection = collection.get("system:id").getInfo()

        # Override the stac argument using the GEE STAC
        stac = "https://earthengine-stac.storage.googleapis.com/catalog/catalog.json"

    else:

        # Convert UTM Bbox to a Feature
        bbox_utm = rasterio.features.bounds(bbox_utm)

        # Open the Catalogue
        CATALOG = pystac_client.Client.open(stac)

        # Do a search
        SEARCH = CATALOG.search(
            intersects=bbox_latlon,
            datetime=f"{start_date}/{end_date}",
            collections=[collection],
            **kwargs,
        )

        # Get all items and sign if using Planetary Computer
        items = SEARCH.item_collection()

        if stac == "https://planetarycomputer.microsoft.com/api/stac/v1":
            items = pc.sign(items)

        # Put the bands into list if not a list already
        if not isinstance(bands, list) and bands is not None:
            bands = [bands]

        # Add stackstac arguments
        if stackstac_kw is None:
            stackstac_kw = dict()

        # Create the cube
        cube = stackstac.stack(
            items,
            assets=bands,
            resolution=resolution,
            bounds=bbox_utm,
            epsg=epsg,
            **stackstac_kw,
        )

        # Delete attributes
        attributes = ["spec", "crs", "transform", "resolution"]

        for attribute in attributes:
            if attribute in cube.attrs:
                del cube.attrs[attribute]

    # Rounded edge size
    rounded_edge_size = cube.x.shape[0]

    # New attributes
    cube.attrs = dict(
        collection=collection,
        stac=stac,
        epsg=epsg,
        resolution=resolution,
        edge_size=rounded_edge_size,
        edge_size_m=rounded_edge_size * resolution,
        central_lat=lat,
        central_lon=lon,
        central_y=utm_coords[1],
        central_x=utm_coords[0],
        time_coverage_start=start_date,
        time_coverage_end=end_date,
    )

    cube = cube.assign_coords(
        {"cubo:distance_from_center": (["y", "x"], _compute_distance_to_center(cube))}
    )

    # New name
    cube.name = collection

    return cube



from typing import Any, List, Union

import numpy as np
import xarray as xr
from pyproj import CRS, Transformer
from pyproj.aoi import AreaOfInterest
from pyproj.database import query_utm_crs_info


def _central_pixel_bbox(
    lat: Union[float, int],
    lon: Union[float, int],
    edge_size: Union[float, int],
    resolution: Union[float, int],
) -> tuple:
    """Creates a Bounding Box (BBox) given a pair of coordinates and a buffer distance.

    Parameters
    ----------
    lat : float
        Latitude.
    lon : float
        Longitude.
    edge_size : float
        Buffer distance in meters.
    resolution : int | float
        Spatial resolution to use.
    latlng : bool, default = True
        Whether to return the BBox as geographic coordinates.

    Returns
    -------
    tuple
        BBox in UTM coordinates, BBox in latlon, and EPSG.
    """
    # Get the UTM EPSG from latlon
    utm_crs_list = query_utm_crs_info(
        datum_name="WGS 84",
        area_of_interest=AreaOfInterest(lon, lat, lon, lat),
    )

    # Save the CRS
    epsg = utm_crs_list[0].code
    utm_crs = CRS.from_epsg(epsg)

    # Initialize a transformer to UTM
    transformer = Transformer.from_crs("EPSG:4326", f"EPSG:{epsg}", always_xy=True)

    # Initialize a transformer from UTM to latlon
    inverse_transformer = Transformer.from_crs(
        f"EPSG:{epsg}", "EPSG:4326", always_xy=True
    )

    # Transform latlon to UTM
    utm_coords = transformer.transform(lon, lat)

    # Round the coordinates
    utm_coords_round = [round(coord / resolution) * resolution for coord in utm_coords]

    # Buffer size
    buffer = round(edge_size * resolution / 2)

    # Create BBox coordinates according to the edge size
    E = utm_coords_round[0] + buffer
    W = utm_coords_round[0] - buffer
    N = utm_coords_round[1] + buffer
    S = utm_coords_round[1] - buffer

    # Create polygon from BBox coordinates
    polygon = [
        [W, S],
        [E, S],
        [E, N],
        [W, N],
        [W, S],
    ]

    # Transform vertices of polygon to latlon
    polygon_latlon = [list(inverse_transformer.transform(x[0], x[1])) for x in polygon]

    # Create UTM BBox
    bbox_utm = {
        "type": "Polygon",
        "coordinates": [polygon],
    }

    # Create latlon BBox
    bbox_latlon = {
        "type": "Polygon",
        "coordinates": [polygon_latlon],
    }

    return (bbox_utm, bbox_latlon, utm_coords, int(epsg))


def _compute_distance_to_center(da: xr.DataArray) -> np.ndarray:
    """Computes the distance from each pixel to the specified center of the data cube.

    Parameters
    ----------
    da : xr.DataArray
        Data cube to compute the distance from.

    Returns
    -------
    np.ndarray
        Distance from each pixel to the specified center.
    """
    # Create meshgrid of coordinates
    coordinates = np.meshgrid(da.x, da.y)

    # Create meshgrid using just the value of the center coordinates
    x = (coordinates[0] ** 0) * da.attrs["central_x"]
    y = (coordinates[1] ** 0) * da.attrs["central_y"]

    # Compute the distance, transposed, so y is first
    distance_to_center = np.linalg.norm((coordinates) - np.array([x, y]), axis=0).T

    return distance_to_center



import unittest

import numpy as np
import pandas as pd
import xarray as xr

import cubo


class Test(unittest.TestCase):
    """Tests for the cubo package."""

    def test_planetary_computer(self):
        """Test the cubo"""
        da = cubo.create(
            lat=50,
            lon=10,
            collection="sentinel-2-l2a",
            bands=["B02", "B03", "B04"],
            start_date="2021-06-01",
            end_date="2021-06-10",
            edge_size=32,
            resolution=10,
        )
        self.assertIsInstance(da, xr.DataArray)

    def test_stackstac(self):
        """Test the cubo with stackstac kwargs"""
        da = cubo.create(
            lat=50,
            lon=10,
            collection="sentinel-2-l2a",
            bands=["B02", "B03", "B04"],
            start_date="2021-06-01",
            end_date="2021-06-10",
            edge_size=32,
            resolution=10,
            stackstac_kw=dict(xy_coords="center"),
        )
        self.assertIsInstance(da, xr.DataArray)

    # def test_element84(self):
    #    """Test the cubo"""
    #    da = cubo.create(
    #        lat=50,
    #        lon=10,
    #        collection="sentinel-s2-l2a-cogs",
    #        bands=["B02","B03","B04"],
    #        start_date="2021-06-01",
    #        end_date="2021-06-10",
    #        edge_size=32,
    #        resolution=10,
    #        stac="https://earth-search.aws.element84.com/v0"
    #    )
    #    self.assertIsInstance(da, xr.DataArray)


if __name__ == "__main__":
    unittest.main()



import io
import os
import re

from setuptools import find_packages, setup


def read(filename):
    filename = os.path.join(os.path.dirname(__file__), filename)
    text_type = type(u"")
    with io.open(filename, mode="r", encoding="utf-8") as fd:
        return re.sub(text_type(r":[a-z]+:`~?(.*?)`"), text_type(r"``\1``"), fd.read())


setup(
    name="cubo",
    version="2024.8.0",
    url="https://github.com/davemlz/cubo",
    license="MIT",
    author="David Montero Loaiza",
    author_email="dml.mont@gmail.com",
    description="On-Demand Earth System Data Cubes in Python",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=("tests",)),
    package_data={"cubo": ["data/*.json"]},
    install_requires=[
        "dask>=2023.7.0",
        "numpy<2.0.0",
        "pandas>=2.0.3",        
        "planetary_computer>=1.0.0",
        "pystac_client>=0.7.2",
        "scipy>=1.12.0",
        "stackstac>=0.4.4",
        "xarray>=2023.6.0",
    ],
    extras_require={
        "ee": ["xee>=0.0.6"],
    },
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)
